﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Linq.Expressions;
using System.Reflection;

namespace ChainOfCommandSample
{
    /// <summary>
    /// Simple implementation of the chain of command.
    /// </summary>
    public static class CommandChainFactory
    {
        private static readonly BindingFlags _methodFlags = 
            BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance;

        /// <summary>
        /// Creates a command chain using reflection.
        /// </summary>
        /// <returns></returns>
        public static List<Command> Create()
        {
            List<Command> commands = new List<Command>();

            Type[] types = Assembly.GetExecutingAssembly().GetTypes();
            Type commandType = typeof (Command);

            Dictionary<string, List<EventCoupling>> eventSources
                = new Dictionary<string, List<EventCoupling>>();

            Dictionary<string, List<MethodCoupling>> eventTargets
                = new Dictionary<string, List<MethodCoupling>>();

            foreach(Type type in types)
            {
                if(!type.IsSubclassOf(commandType))
                {
                    continue;
                }

                MemberInitExpression init = Expression.MemberInit(
                    Expression.New(type), new MemberBinding[0]);

                Command command = Expression.Lambda<Func<Command>>(init)
                    .Compile().Invoke();

                commands.Add(command);

                BuildHandlerCache(command, eventTargets, 
                    type.GetMethods(_methodFlags));

                // Parse the events.
                EventInfo[] events = type.GetEvents(
                    BindingFlags.Public | BindingFlags.Instance);

                if(events.Length == 0)
                {
                    continue; 
                }

                BuildEventCache(command, eventSources, events);
            }

            WireEvents(eventSources, eventTargets);

            commands = commands.OrderBy(c => c.Priority).ToList();

            return commands;
        }

        /// <summary>
        /// Wires the events.
        /// </summary>
        /// <param name="eventSources">The event sources.</param>
        /// <param name="eventTargets">The event targets.</param>
        private static void WireEvents(Dictionary<string, List<EventCoupling>> eventSources, Dictionary<string, List<MethodCoupling>> eventTargets)
        {
            foreach(string key in eventSources.Keys)
            {
                if(!eventTargets.ContainsKey(key))
                {
                    continue;
                }

                List<MethodCoupling> targets = eventTargets[key];
                List<EventCoupling> sources = eventSources[key];

                foreach(EventCoupling source in sources)
                {
                    foreach(MethodCoupling target in targets)
                    {
                        Delegate d = Delegate.CreateDelegate(
                            source.Event.EventHandlerType, target.Command, target.Method);

                        source.Event.AddEventHandler(source.Command, d);
                    }
                }
            }
        }

        /// <summary>
        /// Builds the handler cache.
        /// </summary>
        /// <param name="command">The command.</param>
        /// <param name="eventTargets">The event targets.</param>
        /// <param name="methods">The methods.</param>
        private static void BuildHandlerCache(
            Command command, 
            IDictionary<string, List<MethodCoupling>> eventTargets, 
            IEnumerable<MethodInfo> methods)
        {
            foreach(MethodInfo method in methods)
            {
                EventSubscriberAttribute[] subscriberAttributes =
                    (EventSubscriberAttribute[])
                    method.GetCustomAttributes(typeof(EventSubscriberAttribute), false);

                if (subscriberAttributes.Length == 0)
                {
                    continue;
                }

                foreach(EventSubscriberAttribute attribute in subscriberAttributes)
                {
                    if(!eventTargets.ContainsKey(attribute.EventName))
                    {
                        eventTargets[attribute.EventName] = new List<MethodCoupling>();
                    }

                    eventTargets[attribute.EventName].Add(new MethodCoupling(method, command));
                }
            }
        }

        /// <summary>
        /// Builds the event caches.
        /// </summary>
        /// <param name="command">The command.</param>
        /// <param name="eventSources">The event sources.</param>
        /// <param name="events">The events.</param>
        private static void BuildEventCache(
            Command command,
            IDictionary<string, List<EventCoupling>> eventSources,  
            IEnumerable<EventInfo> events)
        {
            foreach(EventInfo eventInfo in events)
            {
                EventPublisherAttribute[] publisherAttributes =
                    (EventPublisherAttribute[])
                    eventInfo.GetCustomAttributes(typeof (EventPublisherAttribute), false);

                if(publisherAttributes.Length == 0)
                {
                    continue;
                }

                foreach (EventPublisherAttribute attribute in publisherAttributes)
                {
                    if(!eventSources.ContainsKey(attribute.EventName))
                    {
                        eventSources[attribute.EventName] = new List<EventCoupling>();
                    }

                    eventSources[attribute.EventName].Add(new EventCoupling(eventInfo, command));
                }
            }
        }

        /// <summary>
        /// Simple struct to hold a coupling between an event name and event source.
        /// </summary>
        private struct EventCoupling
        {
            private EventInfo _event;
            private Command _command;

            /// <summary>
            /// Gets or sets the event.
            /// </summary>
            /// <value>The event.</value>
            public EventInfo Event
            {
                get { return _event; }
                set { _event = value; }
            }

            /// <summary>
            /// Gets or sets the source.
            /// </summary>
            /// <value>The source.</value>
            public Command Command
            {
                get { return _command; }
                set { _command = value; }
            }

            /// <summary>
            /// Initializes a new instance of the <see cref="EventCoupling"/> struct.
            /// </summary>
            /// <param name="event">The event.</param>
            /// <param name="command">The source.</param>
            public EventCoupling(EventInfo @event, Command command)
            {
                _event = @event;
                _command = command;
            }
        }

        /// <summary>
        /// Simple struct to hold a mapping between a method and a command instance.
        /// </summary>
        private struct MethodCoupling
        {
            private MethodInfo _method;
            private Command _command;

            /// <summary>
            /// Gets or sets the method.
            /// </summary>
            /// <value>The method.</value>
            public MethodInfo Method
            {
                get { return _method; }
                set { _method = value; }
            }

            /// <summary>
            /// Gets or sets the command.
            /// </summary>
            /// <value>The command.</value>
            public Command Command
            {
                get { return _command; }
                set { _command = value; }
            }

            /// <summary>
            /// Initializes a new instance of the <see cref="MethodCoupling"/> struct.
            /// </summary>
            /// <param name="method">The method.</param>
            /// <param name="command">The command.</param>
            public MethodCoupling(MethodInfo method, Command command)
            {
                _method = method;
                _command = command;
            }
        }
    }
}